#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import re
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import List

QUEUE_TIMEOUT = 1  # type: int

RE_IGNORE_PATH = re.compile('^$')  # type: re.Pattern

options = None  # here to include 'options' in globals
options = globals().copy()  # THIS MUST PRECEDE DIRECTLY BEFORE LIST OF CONFIG OPTIONS!

service_name = os.getenv('AGENT_NAME') or 'Python Service Name'  # type: str
zipkin_host = os.getenv('ZIPKIN_HOST') or 'localhost'
zipkin_port = os.getenv('ZIPKIN_PORT') or 9411
force_tls = os.getenv('AGENT_FORCE_TLS', '').lower() == 'true'  # type: bool
protocol = (os.getenv('AGENT_PROTOCOL') or 'grpc').lower()  # type: str
authentication = os.getenv('AGENT_AUTHENTICATION')  # type: str
logging_level = os.getenv('AGENT_LOGGING_LEVEL') or 'INFO'  # type: str
disable_plugins = (os.getenv('AGENT_DISABLE_PLUGINS') or '').split(',')  # type: List[str]
max_buffer_size = int(os.getenv('AGENT_MAX_BUFFER_SIZE', '1000'))  # type: int
ignore_suffix = os.getenv('IGNORE_SUFFIX') or '.jpg,.jpeg,.js,.css,.png,.bmp,.gif,.ico,.mp3,' \
                                                 '.mp4,.html,.svg '  # type: str
flask_collect_http_params = True if os.getenv('FLASK_COLLECT_HTTP_PARAMS') and \
                                    os.getenv('FLASK_COLLECT_HTTP_PARAMS') == 'True' else False  # type: bool
http_params_length_threshold = int(os.getenv('HTTP_PARAMS_LENGTH_THRESHOLD') or '1024')  # type: int
trace_ignore_path = os.getenv('TRACE_IGNORE_PATH') or ''  # type: str
profile_active = True if os.getenv('AGENT_PROFILE_ACTIVE') and \
                         os.getenv('AGENT_PROFILE_ACTIVE') == 'True' else False  # type: bool
profile_task_query_interval = int(os.getenv('PROFILE_TASK_QUERY_INTERVAL') or '20')

options = {key for key in globals() if key not in options}  # THIS MUST FOLLOW DIRECTLY AFTER LIST OF CONFIG OPTIONS!


def init(**kwargs):
    glob = globals()

    for key, val in kwargs.items():
        if key not in options:
            raise KeyError('invalid config option %s' % key)

        glob[key] = val



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import pathlib

from setuptools import setup, find_packages

HERE = pathlib.Path(__file__).parent

README = (HERE / "README.md").read_text()

setup(
    name="zipkin-agent",
    version="0.1.0",
    description="Python Agent for zipkin-agent",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/apache/skywalking-python/",
    author="Apache",
    author_email="dev@skywalking.apache.org",
    license="Apache 2.0",
    packages=find_packages(exclude=("tests", "tests.*")),
    include_package_data=True,
    install_requires=[
        "packaging",
        "wrapt",
    ],
    extras_require={
        "test": [
            "testcontainers",
            "pyyaml",
            "pytest",
        ],
        "http": [
            "requests>=2.26.0",
        ],
    },
    classifiers=[
        "Framework :: Flake8",

        "License :: OSI Approved :: Apache Software License",

        "Operating System :: OS Independent",

        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",

        "Topic :: Software Development",
    ]
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# @Time    : 2021/11/13 3:21 上午
# @Author  : chenxuhan  
# @File    : sw_dapr.py
# @Software: PyCharm


from opencensus.trace.execution_context import get_opencensus_tracer, get_current_span
from opencensus.trace.tracers.noop_tracer import NoopTracer


def install():
    from dapr.clients import DaprClient
    from typing import Optional, Union
    from dapr.clients.grpc.client import MetadataTuple, InvokeMethodResponse
    from google.protobuf.message import Message as GrpcMessage

    _invoke_method = DaprClient.invoke_method

    def _zk_invoke_method(
            this: DaprClient,
            app_id: str,
            method_name: str,
            data: Union[bytes, str, GrpcMessage],
            content_type: Optional[str] = None,
            metadata: Optional[MetadataTuple] = None,
            http_verb: Optional[str] = None,
            http_querystring: Optional[MetadataTuple] = None) -> InvokeMethodResponse:

        _tracer = get_opencensus_tracer()
        if not isinstance(_tracer, NoopTracer):
            if metadata is None:
                metadata = []
            metadata.extend([(key, val) for key, val in _tracer.propagator.to_headers(_tracer.span_context).items()])
        res = _invoke_method(this, app_id, method_name, data, content_type, metadata, http_verb, http_querystring)

        return res

    DaprClient.invoke_method = _zk_invoke_method



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from zipkin_agent import config, plugins, loggings
from opencensus.trace.tracer import Tracer
from opencensus.ext.zipkin.trace_exporter import ZipkinExporter
from opencensus.trace.samplers import AlwaysOnSampler
from opencensus.trace.execution_context import set_opencensus_tracer


__started = False


def __init():
    ze = ZipkinExporter(
        service_name=config.service_name,
        host_name=config.zipkin_host,
        port=config.zipkin_port,
        endpoint='/api/v2/spans')
    setattr(config, 'zipkin_reporter', ze)

    plugins.install()


def start():
    global __started
    if __started:
        return
    __started = True

    loggings.init()

    __init()


def started():
    return __started


def init_zk_tracer(**kwargs):
    _tracer = Tracer(
        exporter=getattr(config, 'zipkin_reporter'),
        sampler=AlwaysOnSampler(),
        **kwargs
    )
    set_opencensus_tracer(_tracer)


#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from zipkin_agent import Layer, Component, config
from zipkin_agent.agent import init_zk_tracer
from zipkin_agent.trace import tags

from opencensus.trace.propagation.trace_context_http_header_format import TraceContextPropagator
from opencensus.trace.status import Status
from opencensus.trace.execution_context import get_opencensus_tracer, get_current_span


def install():
    from flask import Flask
    _full_dispatch_request = Flask.full_dispatch_request
    _handle_user_exception = Flask.handle_user_exception
    _handle_exception = Flask.handle_exception

    def params_tostring(params):
        return "\n".join([k + '=[' + ",".join(params.getlist(k)) + ']' for k, _ in params.items()])

    def _zk_full_dispatch_request(this: Flask):
        import flask
        req = flask.request

        if 'traceparent' in req.headers:
            init_zk_tracer(span_context=TraceContextPropagator().from_headers(req.headers))
            resp = _full_dispatch_request(this)
            return resp
        else:
            init_zk_tracer()
            with get_opencensus_tracer().span(name=req.url.split("?")[0]) as span:
                span.add_attribute('kind', 'client')
                span.add_attribute('layer', Layer.Http.name)
                span.add_attribute('component', Component.Flask.name)
                span.add_attribute('peer', '%s:%s' % (req.environ["REMOTE_ADDR"], req.environ["REMOTE_PORT"]))
                span.add_attribute(tags.HttpMethod, req.method)
                span.add_attribute(tags.HttpUrl, req.url.split("?")[0])
                span.add_attribute(tags.HttpParams, params_tostring(req.values)[0:config.http_params_length_threshold])
                resp = _full_dispatch_request(this)
                span.set_status(Status(code=resp.status_code))
                span.add_attribute(tags.HttpStatus, resp.status_code)
                return resp

    def _zk_handle_user_exception(this: Flask, e):
        if e is not None:
            entry_span = get_current_span()
            if entry_span is not None:
                entry_span.set_status(Status.from_exception(e))

        return _handle_user_exception(this, e)

    def _zk_handle_exception(this: Flask, e):
        if e is not None:
            entry_span = get_current_span()
            if entry_span is not None:
                entry_span.set_status(Status.from_exception(e))

        return _handle_exception(this, e)


    Flask.full_dispatch_request = _zk_full_dispatch_request
    Flask.handle_user_exception = _zk_handle_user_exception
    Flask.handle_exception = _zk_handle_exception


#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from collections import namedtuple

Tag = namedtuple('Tag', 'key val overridable')
Tag.__new__.__defaults__ = (None, None, False)

HttpUrl = 'url'
HttpMethod = 'http.method'
HttpStatus = 'status.code'
DbType = 'db.type'
DbInstance = 'db.instance'
DbStatement = 'db.statement'
DbSqlParameters = 'db.sql.parameters'
HttpParams = 'http.params'
MqBroker = 'mq.broker'
MqTopic = 'mq.topic'
MqQueue = 'mq.queue'
CeleryParameters = 'celery.parameters'



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from functools import wraps
import inspect

from zipkin_agent import Layer, Component
from opencensus.trace.execution_context import get_opencensus_tracer


def trace(
        op: str = None,
        layer: Layer = Layer.Unknown,
        component: Component = Component.Unknown,
        tags: dict = None,
):
    def decorator(func):
        _op = op or func.__name__

        if inspect.iscoroutinefunction(func):
            @wraps(func)
            async def wrapper(*args, **kwargs):
                _tracer = get_opencensus_tracer()
                span = _tracer.span(name=_op)
                span.add_attribute('layer', layer.name)
                span.add_attribute('component', component.name)
                if tags:
                    for tag_k, tag_v in tags.items():
                        span.add_attribute(tag_k, tag_v)
                with span:
                    return await func(*args, **kwargs)
            return wrapper

        else:
            @wraps(func)
            def wrapper(*args, **kwargs):
                _tracer = get_opencensus_tracer()
                span = _tracer.span(name=_op)
                span.add_attribute('layer', layer.name)
                span.add_attribute('component', component.name)
                if tags:
                    for tag_k, tag_v in tags.items():
                        span.add_attribute(tag_k, tag_v)
                with span:
                    return func(*args, **kwargs)
            return wrapper

    return decorator


def runnable(
        op: str = None,
        layer: Layer = Layer.Unknown,
        component: Component = Component.Unknown,
        tags: dict = None,
):
    def decorator(func):

        @wraps(func)
        def wrapper(*args, **kwargs):
            _op = op or "Thread/"+func.__name__
            _tracer = get_opencensus_tracer()
            with _tracer.new_local_span(name=_op) as span:
                span.add_attribute('layer', layer.name)
                span.add_attribute('component', component.name)
                if tags:
                    for tag_k, tag_v in tags.items():
                        span.add_attribute(tag_k, tag_v)
                func(*args, **kwargs)

        return wrapper

    return decorator



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#


from opencensus.trace.execution_context import get_opencensus_tracer
from opencensus.trace.tracers.noop_tracer import NoopTracer


def install():
    from requests import Session

    _request = Session.request

    def _zk_request(this: Session, method, url,
                    params=None, data=None, headers=None, cookies=None, files=None,
                    auth=None, timeout=None, allow_redirects=True, proxies=None,
                    hooks=None, stream=None, verify=None, cert=None, json=None):

        from urllib.parse import urlparse
        url_param = urlparse(url)

        _tracer = get_opencensus_tracer()

        # ignore trace reporter self request
        if isinstance(_tracer, NoopTracer) or _tracer.exporter.get_url.endswith(url_param.netloc + url_param.path):
            return _request(this, method, url, params, data, headers, cookies, files, auth, timeout,
                            allow_redirects,
                            proxies,
                            hooks, stream, verify, cert, json)

        if not isinstance(_tracer, NoopTracer):
            if headers is None:
                headers = {}
            trace_headers = _tracer.propagator.to_headers(_tracer.span_context)
            for key in trace_headers:
                headers[key] = trace_headers[key]
        res = _request(this, method, url, params, data, headers, cookies, files, auth, timeout,
                       allow_redirects,
                       proxies,
                       hooks, stream, verify, cert, json)

        return res

    Session.request = _zk_request



#
# Licensed to the Apache Software Foundation (ASF) under one or more
# contributor license agreements.  See the NOTICE file distributed with
# this work for additional information regarding copyright ownership.
# The ASF licenses this file to You under the Apache License, Version 2.0
# (the "License"); you may not use this file except in compliance with
# the License.  You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import inspect
import logging
from zipkin_agent.loggings import logger
import pkgutil
import re
import traceback

import pkg_resources

from packaging import version

from zipkin_agent import config

import zipkin_agent


def install():
    disable_patterns = config.disable_plugins
    if isinstance(disable_patterns, str):
        disable_patterns = [re.compile(p.strip()) for p in disable_patterns.split(',') if p.strip()]
    else:
        disable_patterns = [re.compile(p.strip()) for p in disable_patterns if p.strip()]
    for importer, modname, ispkg in pkgutil.iter_modules(zipkin_agent.plugins.__path__):
        if any(pattern.match(modname) for pattern in disable_patterns):
            logger.info('plugin %s is disabled and thus won\'t be installed', modname)
            continue
        logger.debug('installing plugin %s', modname)
        plugin = importer.find_module(modname).load_module(modname)

        supported = pkg_version_check(plugin)
        if not supported:
            logger.debug('check version for plugin %s\'s corresponding package failed, thus '
                         'won\'t be installed', modname)
            continue

        if not hasattr(plugin, 'install') or inspect.ismethod(getattr(plugin, 'install')):
            logger.warning('no `install` method in plugin %s, thus the plugin won\'t be installed', modname)
            continue

        # noinspection PyBroadException
        try:
            plugin.install()
        except Exception:
            logger.warning('failed to install plugin %s', modname)
            traceback.print_exc() if logger.isEnabledFor(logging.DEBUG) else None


_operators = {
    '<': lambda cv, ev: cv < ev,
    '<=': lambda cv, ev: cv < ev or cv == ev,
    '==': lambda cv, ev: cv == ev,
    '>=': lambda cv, ev: cv > ev or cv == ev,
    '>': lambda cv, ev: cv > ev,
    '!=': lambda cv, ev: cv != ev
}


class VersionRuleException(Exception):
    def __init__(self, message):
        self.message = message


def pkg_version_check(plugin):
    supported = True

    # no version rules was set, no checks
    if not hasattr(plugin, "version_rule"):
        return supported

    pkg_name = plugin.version_rule.get("name")
    rules = plugin.version_rule.get("rules")

    try:
        current_pkg_version = pkg_resources.get_distribution(pkg_name).version
    except pkg_resources.DistributionNotFound:
        # when failed to get the version, we consider it as supported.
        return supported

    current_version = version.parse(current_pkg_version)
    # pass one rule in rules (OR)
    for rule in rules:
        if rule.find(" ") == -1:
            if check(rule, current_version):
                return supported
        else:
            # have to pass all rule_uint in this rule (AND)
            rule_units = rule.split(" ")
            results = [check(unit, current_version) for unit in rule_units]
            if False in results:
                # check failed, try to check next rule
                continue
            else:
                return supported

    supported = False
    return supported


def check(rule_unit, current_version):
    idx = 2 if rule_unit[1] == '=' else 1
    symbol = rule_unit[0:idx]
    expect_pkg_version = rule_unit[idx:]

    expect_version = version.parse(expect_pkg_version)
    f = _operators.get(symbol) or None
    if not f:
        raise VersionRuleException("version rule {} error. only allow >,>=,==,<=,<,!= symbols".format(rule_unit))

    return f(current_version, expect_version)
